;; -*- mode: Scheme; -*-
;;
;; This file is part of SRFI-99 for CHICKEN
;; Copyright (c) 2011 by Thomas Chust.  All rights reserved.
;;
;; Permission is hereby granted, free of charge, to any person
;; obtaining a copy of this software and associated documentation
;; files (the Software), to deal in the Software without restriction,
;; including without limitation the rights to use, copy, modify,
;; merge, publish, distribute, sublicense, and/or sell copies of the
;; Software, and to permit persons to whom the Software is furnished
;; to do so, subject to the following conditions:
;; 
;; The above copyright notice and this permission notice shall be
;; included in all copies or substantial portions of the Software.
;; 
;; THE SOFTWARE IS PROVIDED ASIS, WITHOUT WARRANTY OF ANY KIND,
;; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
;; MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
;; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
;; BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
;; ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
;; CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
;; SOFTWARE.

(require-library
 srfi-1 srfi-69
 data-structures)

(module srfi-99-primitives
  (%make-rtd %get-rtd rtd?
   %rtd-name %rtd-uid %rtd-child-uids %rtd-fields %rtd-parent
   %rtd-child-uid?
   %rtd-count-fields %rtd-count-all-fields
   %rtd-field-ref %rtd-field-find)
  (import
   scheme chicken
   srfi-69)

(define-values (%make-rtd %get-rtd)
  (letrec ((%link!
	    (lambda (rtd)
	      (do ((rtd rtd (%rtd-parent rtd)) (uid (%rtd-uid rtd))) ((not rtd))
		(cond
		 ((%rtd-child-uids rtd) => (cut hash-table-set! <> uid #t))))))
	   (%unlink!
	    (lambda (rtd)
	      (do ((rtd rtd (%rtd-parent rtd)) (uid (%rtd-uid rtd))) ((not rtd))
		(cond
		 ((%rtd-child-uids rtd) => (cut hash-table-delete! <> uid))))))
	   (%make-rtd
	    (lambda (name uid fields parent sealed? opaque?)
	      (let ((rtd (##sys#make-structure
			  'rtd
			  name uid
			  (and (not sealed?)
			       (make-hash-table #:test eq? #:hash eq?-hash))
			  fields parent)))
		(set-finalizer! rtd %unlink!)
		(%link! rtd)
		(unless opaque? (put! uid 'rtd rtd))
		rtd)))
	   (%get-rtd
	    (lambda (uid)
	      (get uid 'rtd #f))))
    (values %make-rtd %get-rtd)))
  
(define (rtd? v) (##sys#structure? v 'rtd))

(define (%rtd-name rtd) (##sys#slot rtd 1))
(define (%rtd-uid rtd) (##sys#slot rtd 2))
(define (%rtd-child-uids rtd) (##sys#slot rtd 3))
(define (%rtd-fields rtd) (##sys#slot rtd 4))
(define (%rtd-parent rtd) (##sys#slot rtd 5))

(define (%rtd-child-uid? rtd uid)
  (cond
   ((%rtd-child-uids rtd) => (cut hash-table-ref/default <> uid #f))
   (else (eq? uid (%rtd-uid rtd)))))

(define (%rtd-count-fields rtd)
  (vector-length (%rtd-fields rtd)))

(define (%rtd-count-all-fields rtd)
  (let loop ((rtd rtd) (n 0))
    (if rtd
	(loop (%rtd-parent rtd) (+ n (%rtd-count-fields rtd)))
	n)))

(define (%rtd-field-ref rtd i)
  (let ((base (- (%rtd-count-all-fields rtd) (%rtd-count-fields rtd))))
    (if (>= i base)
	(vector-ref (%rtd-fields rtd) (- i base))
	(%rtd-field-ref (%rtd-parent rtd) i))))

(define (%rtd-field-find rtd name)
  (let ((n (%rtd-count-all-fields rtd)))
    (let loop ((i (fx- n 1)))
      (if (negative? i)
	  (error "no such field in record" name (%rtd-name rtd))
	  (let ((field (%rtd-field-ref rtd i)))
	    (if (eq? (cadr field) name)
		(apply values i field)
		(loop (fx- i 1))))))))

)

(module srfi-99-records-procedural
  (make-rtd rtd?
   rtd-constructor rtd-predicate rtd-accessor rtd-mutator)
  (import
   scheme chicken
   srfi-1 srfi-99-primitives data-structures)

(define (make-rtd name fields #!optional parent #!key uid sealed opaque)
  (##sys#check-symbol name 'make-rtd)
  (##sys#check-vector fields 'make-rtd)
  (when parent
    (##sys#check-structure parent 'rtd)
    (ensure %rtd-child-uids parent
	    'make-rtd "expected unsealed parent" parent))
  (when uid
    (##sys#check-symbol uid 'make-rtd))
  (%make-rtd
   name
   (or uid (gensym name))
   (list->vector
    (map
     (lambda (field)
       (cond
	((symbol? field)
	 (list 'immutable field))
	((and (pair? field) (symbol? (car field))
	      (null? (cdr field)))
	 (cons 'mutable field))
	((and (pair? field) (memq (car field) '(mutable immutable))
	      (pair? (cdr field)) (symbol? (cadr field))
	      (null? (cddr field)))
	 field)
	(else
	 (error 'make-rtd "bad field specification" field))))
     (vector->list fields)))
   parent sealed opaque))

(define (rtd-constructor rtd #!optional fields)
  (##sys#check-structure rtd 'rtd 'rtd-constructor)
  (let* ((n (%rtd-count-all-fields rtd))
	 (is (if fields
		 (map
		  (lambda (name)
		    (let-values (((i access field) (%rtd-field-find rtd name)))
		      (fx+ i 1)))
		  (vector->list fields))
		 (iota n 1))))
    (lambda vs
      (unless (= (length vs) (length is))
	(##sys#signal-hook
	 #:arity-error
	 (string-append
	  "bad argument count - received " (number->string (length vs))
	  " but expected " (number->string (length is)))))
      (let ((r (##sys#allocate-vector (fx+ n 1) #f (void) #f)))
	(##sys#setslot r 0 (%rtd-uid rtd))
	(##core#inline "C_vector_to_structure" r)
	(for-each (cut ##sys#setslot r <> <>) is vs)
	r))))

(define (%rtd-predicate rtd)
  (lambda (v)
    (and (##sys#generic-structure? v)
	 (%rtd-child-uid? rtd (##sys#slot v 0)))))

(define (rtd-predicate rtd)
  (##sys#check-structure rtd 'rtd 'rtd-predicate)
  (%rtd-predicate rtd))

(define (rtd-accessor rtd name)
  (##sys#check-structure rtd 'rtd 'rtd-accessor)
  (let-values (((instance?) (%rtd-predicate rtd))
	       ((i access field) (%rtd-field-find rtd name)))
    (set! i (fx+ i 1))
    ((if (eq? access 'mutable)
	 (cut getter-with-setter <>
	      (lambda (r v)
		(ensure instance? r)
		(##sys#setslot r i v)))
	 identity)
     (lambda (r)
       (ensure instance? r)
       (##sys#slot r i)))))

(define (rtd-mutator rtd name)
  (##sys#check-structure rtd 'rtd 'rtd-accessor)
  (##sys#check-structure rtd 'rtd 'rtd-accessor)
  (let-values (((instance?) (%rtd-predicate rtd))
	       ((i access field) (%rtd-field-find rtd name)))
    (set! i (fx+ i 1))
    (and (eq? access 'mutable)
	 (lambda (r v)
	   (ensure instance? r)
	   (##sys#setslot r i v)))))

)

(module srfi-99-records-inspection
  (record? record-rtd
   rtd-name rtd-uid rtd-sealed? rtd-parent
   rtd-field-names rtd-all-field-names rtd-field-mutable?)
  (import
   scheme chicken
   srfi-99-primitives)

(define (record? v)
  (##sys#generic-structure? v))

(define (record-rtd v)
  (and (##sys#generic-structure? v)
       (%get-rtd (##sys#slot v 0))))

(define (rtd-name rtd)
  (##sys#check-structure rtd 'rtd 'rtd-name)
  (%rtd-name rtd))

(define (rtd-uid rtd)
  (##sys#check-structure rtd 'rtd 'rtd-uid)
  (%rtd-uid rtd))

(define (rtd-sealed? rtd)
  (##sys#check-structure rtd 'rtd 'rtd-sealed?)
  (not (%rtd-child-uids rtd)))

(define (rtd-parent rtd)
  (##sys#check-structure rtd 'rtd 'rtd-parent)
  (%rtd-parent rtd))

(define (rtd-field-names rtd)
  (##sys#check-structure rtd 'rtd 'rtd-field-names)
  (let* ((n (%rtd-count-fields rtd))
	 (n-all (%rtd-count-all-fields rtd))
	 (base (- n-all n)))
    (do ((names (make-vector n)) (i base (fx+ i 1))) ((>= i n-all) names)
      (vector-set! names (- i base) (cadr (%rtd-field-ref rtd i))))))

(define (rtd-all-field-names rtd)
  (##sys#check-structure rtd 'rtd 'rtd-all-field-names)
  (let ((n (%rtd-count-fields rtd))
	(n-all (%rtd-count-all-fields rtd)))
    (do ((names (make-vector n-all)) (i 0 (fx+ i 1))) ((>= i n-all) names)
      (vector-set! names i (cadr (%rtd-field-ref rtd i))))))

(define (rtd-field-mutable? rtd name)
  (##sys#check-structure rtd 'rtd 'rtd-field-mutable?)
  (let-values (((i access field) (%rtd-field-find rtd name)))
    (eq? access 'mutable)))

)

(module srfi-99-records-syntactic
  (define-record-type
   define-record-constructor
   %define-record-constructor/default
   define-record-predicate
   %define-record-predicate/default
   define-record-field
   %define-record-field/mutable-default %define-record-field/immutable-default)
  (import
   scheme (except chicken define-record-type)
   srfi-99-records-procedural)

(define-syntax %define-record-constructor/default
  (ir-macro-transformer
   (lambda (stx inject id=)
     (let* ((rtd (cadr stx))
	    (name (symbol-append 'make- (inject rtd))))
       `(define-record-constructor ,rtd ,(inject name))))))

(define-syntax define-record-constructor
  (syntax-rules ()
    ((define-record-constructor rtd #f)
     (begin))
    ((define-record-constructor rtd #t)
     (%define-record-constructor/default rtd))
    ((define-record-constructor rtd (name field ...))
     (define name
       (rtd-constructor rtd (vector 'field ...))))
    ((define-record-constructor rtd name)
     (define name
       (rtd-constructor rtd)))))

(define-syntax %define-record-predicate/default
  (ir-macro-transformer
   (lambda (stx inject id=)
     (let* ((rtd (cadr stx))
	    (name (symbol-append (inject rtd) '?)))
       `(define-record-predicate ,rtd ,(inject name))))))

(define-syntax define-record-predicate
  (syntax-rules ()
    ((define-record-predicate rtd #f)
     (begin))
    ((define-record-predicate rtd #t)
     (%define-record-predicate/default rtd))
    ((define-record-predicate rtd name)
     (define name
       (rtd-predicate rtd)))))

(define-syntax %define-record-field/mutable-default
  (ir-macro-transformer
   (lambda (stx inject id=)
     (let* ((rtd (cadr stx))
	    (name (caddr stx))
	    (accessor (symbol-append (inject rtd) '- (inject name)))
	    (mutator (symbol-append accessor '-set!)))
       `(define-record-field ,rtd (,name ,(inject accessor) ,(inject mutator)))))))

(define-syntax %define-record-field/immutable-default
  (ir-macro-transformer
   (lambda (stx inject id=)
     (let* ((rtd (cadr stx))
	    (name (caddr stx))
	    (accessor (symbol-append (inject rtd) '- (inject name))))
       `(define-record-field ,rtd (,name ,(inject accessor)))))))

(define-syntax define-record-field
  (syntax-rules ()
    ((define-record-field rtd (name accessor mutator))
     (begin
       (define accessor
	 (rtd-accessor rtd 'name))
       (define mutator
	 (rtd-mutator rtd 'name))))
    ((define-record-field rtd (name accessor))
     (define accessor
       (rtd-accessor rtd 'name)))
    ((define-record-field rtd (name))
     (%define-record-field/mutable-default rtd name))
    ((define-record-field rtd name)
     (%define-record-field/immutable-default rtd name))))

(define-syntax define-record-type
  (syntax-rules ()
    ((define-record-type (rtd parent option ...)
       constructor-spec
       predicate-spec
       field-spec ...)
     (begin
       (define rtd
	 (make-rtd
	  'rtd
	  (let-syntax ((%record-field-spec
			(syntax-rules ()
			  ((record-field-spec (name accessor mutator))
			   '(mutable name))
			  ((record-field-spec (name accessor))
			   '(immutable name))
			  ((record-field-spec (name))
			   '(mutable name))
			  ((record-field-spec name)
			   '(immutable name)))))
	    (vector (%record-field-spec field-spec) ...))
	  parent option ...))
       (define-record-constructor rtd constructor-spec)
       (define-record-predicate rtd predicate-spec)
       (define-record-field rtd field-spec) ...))
    ((define-record-type rtd
       constructor-spec
       predicate-spec
       field-spec ...)
     (define-record-type (rtd #f)
       constructor-spec
       predicate-spec
       field-spec ...))))

)

(module srfi-99-records
  ()
  (import
   scheme chicken)
  (reexport
   srfi-99-records-procedural srfi-99-records-inspection
   srfi-99-records-syntactic)

)

(module srfi-99
  ()
  (import
   scheme chicken)
  (reexport
   srfi-99-records)

)
